// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double* m_dimensionattrs;
	double m_measureattrs;
	double m_onedimensionIndex;
	int m_dimension;
	long m_id;

	Point(int dimension)
	{
		m_dimension=dimension;
		m_dimensionattrs=(double*)malloc(dimension*sizeof(double));
	}
	
	void computeOneDimensionIndex(unsigned long domainsize)
	{
		m_onedimensionIndex=0.0;
		for (int i=0; i<m_dimension; i++)
			m_onedimensionIndex+=m_dimensionattrs[i]*pow((double)domainsize, (double)(m_dimension-(i+1)));
	}
	
	~Point()
	{
		//delete[] m_dimensionattrs;
	}
};

int main(int argc, char** argv)
{
	if (argc != 6)
	{
		cerr << "Usage: " << argv[0] << " dimension measureAttr_domainSize dimensionAttr_domainSize density sort[0|1]." << endl;
		return -1;
	}

	int dimension = atoi(argv[1]);
	unsigned long measure_domainsize= atol(argv[2]);
	unsigned long dimension_domainsize = atol(argv[3]);
	double density = atof(argv[4]);
	//if (dimension>1)
		//density=pow(density, 1.0/(double)dimension);
	cerr<<"density "<<density<<endl;
	int sorting=atoi(argv[5]);
	map<unsigned long, Point> data;
	
	Tools::Random rnd;
	
	unsigned long numberOfObjects=(unsigned long)floor(density*pow((double)dimension_domainsize, (double)dimension));
	set<unsigned long> examined;
	vector<unsigned long> keys;

	for (unsigned long i = 0; i < numberOfObjects; i++)
	{
		double x;
		Point r= Point(dimension);
		set<unsigned long>::iterator itSet;
		do 
		{
			for (int j=0; j<dimension; j++)
			{
				x= rnd.nextUniformDouble();
				x=floor(x*dimension_domainsize);
				r.m_dimensionattrs[j]=x;
			}
			x= rnd.nextUniformDouble();
			x=ceil(x*measure_domainsize);
			r.m_measureattrs=x;
			r.computeOneDimensionIndex(dimension_domainsize);
			itSet = examined.find((unsigned long)r.m_onedimensionIndex);
		}while (itSet != examined.end());
		r.m_id=i;
		examined.insert((unsigned long)r.m_onedimensionIndex);
		data.insert(pair<unsigned long, Point>((unsigned long)r.m_onedimensionIndex, r));
		keys.push_back((unsigned long)r.m_onedimensionIndex);
		if (sorting==0)
		{
			cout << INSERT << " " << i << " "<<r.m_measureattrs;
			for (int j=0;j<dimension; j++)
				cout <<" "<< r.m_dimensionattrs[j];
			cout<<" "<<r.m_onedimensionIndex<<endl;
		}
	}
	if (sorting==1)
	{
		Tools::ResourceUsage ru;
		ru.start();
		sort(keys.begin(), keys.end());
		map<unsigned long, Point>::iterator it;
		for (long i=0;i<keys.size();i++)
		{
			it=data.find(keys[i]);
			cout << INSERT << " " << (*it).second.m_id << " "<<(*it).second.m_measureattrs;
			for (int j=0;j<dimension; j++)
				cout <<" "<< (*it).second.m_dimensionattrs[j];
			cout<<" "<<(*it).second.m_onedimensionIndex<<endl;
		}
		ru.stop();
		cerr << "Seconds: " << ru.getTotalTime() << endl;
		//for (unsigned long i = 0; i < numberOfObjects; i++)
			//cout << INSERT << " " << i << " " << keys[i] << " " << keys[i] << endl;
	}
	return 0;
}
