// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

int main(int argc, char** argv)
{
	if (argc != 6)
	{
		cerr << "Usage: " << argv[0] << " number_of_query selectivity domainsize dimension density." << endl;
		return -1;
	}

	long numberOfQuery = atol(argv[1]);
	double selectivity= atof(argv[2]);
	long domainsize=atol(argv[3]);
	int dimension=atoi(argv[4]);
	double density= atof(argv[5]);
	
	double oneDdensity=pow(density, 1.0/(double)dimension);
	double oneDselectivity=pow(selectivity, 1.0/(double)dimension);
	cerr<<"one dimension density: "<<oneDdensity<<endl;
	cerr<<"one dimension selectivity: "<<oneDselectivity<<endl;

	Tools::Random rnd;

	
	long range=(long)ceil(oneDselectivity*oneDdensity*domainsize);
	long* keys=new long[dimension];
	long* keye=new long[dimension];
	for (long i=0; i< numberOfQuery; i++)
	{
		for (int j=0;j<dimension;j++)
		{
			keys[j] = static_cast<long>(rnd.nextUniformLong(0, domainsize-range));
			keye[j] = keys[j]+range;
		}
		cout << QUERY << " ";
		for (int j=0;j<dimension;j++)
			cout<<keys[j]<<" ";
		for (int j=0;j<dimension;j++)
			cout<<keye[j]<<" ";
		cout<<endl;
	}
	return 0;
}
