// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double* m_dimensionattrs;
	double m_measureattrs;
	double m_onedimensionIndex;
	int m_dimension;

	Point(int dimension)
	{
		m_dimension=dimension;
		m_dimensionattrs=(double*)malloc(dimension*sizeof(double));
	}
	
	void computeOneDimensionIndex(unsigned long domainsize)
	{
		m_onedimensionIndex=0.0;
		for (int i=0; i<m_dimension; i++)
			m_onedimensionIndex+=m_dimensionattrs[i]*pow((double)domainsize, (double)(m_dimension-(i+1)));
	}
	
	~Point()
	{
		//delete[] m_dimensionattrs;
	}
};

int main(int argc, char** argv)
{
	if (argc != 6)
	{
		cerr << "Usage: " << argv[0] << " data_file number_of_query selectivity domainsize dimension." << endl;
		return -1;
	}

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open data file " << argv[1] << "." << endl;
		return -1;
	}

	long numberOfQuery = atol(argv[2]);
	double selectivity= atof(argv[3]);
	long domainsize=atol(argv[4]);
	int dimension=atoi(argv[5]);
	map<long, Point> data;
	vector<long*> keys;
	Tools::Random rnd;
	int op;
	long id;

	while (fin)
	{
		fin >> op >> id;
		Point p(dimension);
		fin >> p.m_measureattrs;
		for (int i=0;i<dimension;i++)
			fin >> p.m_dimensionattrs[i];
		fin >> p.m_onedimensionIndex;
		if (! fin.good()) continue; // skip newlines, etc.
		data.insert(pair<long, Point>(id, p));	
		long* coords=(long*)malloc(dimension*sizeof(long));
		for (int i=0;i<dimension;i++)
			coords[i]=(long)(p.m_dimensionattrs[i]);
		keys.push_back(coords);
	}
	
	//std::sort(keys.begin(),keys.end());

	long numberOfObjects=keys.size();
	long range=(long)ceil(selectivity*numberOfObjects);
	for (long i = 0; i < numberOfQuery; i++)
	{
		long start = static_cast<long>(rnd.nextUniformLong(0, numberOfObjects-range-1));
		/*long x1 = keys[start];
		long x2 = keys[start+range-2];
		map<long, Point>::iterator it1 = data.find(x1);
		assert(it1 != data.end());
		map<long, Point>::iterator it2 = data.find(x2);
		assert(it2 != data.end());*/
		cout << QUERY << " ";
		for (int i=0;i<dimension;i++)
			cout<<keys[start][i]<<" ";
		for (int i=0;i<dimension;i++)
			cout<<keys[start+range-1][i]<<" ";
		cout<<endl;

	}
	return 0;
}
