// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

// NOTE: Please read README.txt before browsing this code.

// include library header file.
#include <SpatialIndex.h>

using namespace SpatialIndex;
using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2

// example of a Visitor pattern.
// findes the index and leaf IO for answering the query and prints
// the resulting data IDs to stdout.
class MyVisitor : public IVisitor
{
public:
	unsigned long m_indexIO;
	unsigned long m_leafIO;
	unsigned long m_affectedNode;
	unsigned long m_mcnSize;
	unsigned long m_mcnVOSize;
	unsigned long m_leafmcn;
	unsigned long m_indexmcn;

	vector<byte* > vos;	
	vector<long> aggs;
	vector<long > digestSize;
	vector<long> digPosition;
	vector<long> mcnAggs;
	vector<byte*> mcnDigest;
	vector<double*>mcnMBRs;
	vector<long> mcnVOaggs;
	vector<byte*> mcnVOdigest;
	vector<double*> mcnVOMBRs;
	vector<long> mcnAffectedNode;
	long m_agg;
	int m_dimension;


public:
	MyVisitor() : m_indexIO(0), m_leafIO(0), m_leafmcn(0),m_indexmcn(0) {}

	void visitNode(const INode& n)
	{
		if (n.isLeaf()) m_leafIO++;
		else m_indexIO++;
	}

	void visitData(const IData& d)
	{
		IShape* pS;
		d.getShape(&pS);
			// do something.
		delete pS;

		// data should be an array of characters representing a Region as a string.
		//byte* pData = 0;
		//unsigned long cLen = 0;
		//d.getData(cLen, &pData);
			// do something.
			//string s = reinterpret_cast<char*>(pData);
			//cout << s << endl;
		//delete[] pData;

		cout << d.getIdentifier() << endl;
			// the ID of this data entry is an answer to the query. I will just print it to stdout.
	}

	void visitData(std::vector<const IData*>& v)
	{
		cout << v[0]->getIdentifier() << " " << v[1]->getIdentifier() << endl;
	}
	
	void updateVOs(byte** digest, long* agg, long num, long pos) 
	{
		digestSize.insert(digestSize.begin(),num);
		digPosition.insert(digPosition.begin(),pos);
		for (long i=num-1;i>=0;i--)
		{
			byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
			memcpy(dig, digest[i], DIGEST_LENGTH);
			vos.insert(vos.begin(),dig);	
			aggs.insert(aggs.begin(),agg[i]);
		}		
	}
	
	void updateMCN(long agg, byte* digest)
	{
		byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		memcpy(dig, digest, DIGEST_LENGTH);
		mcnDigest.insert(mcnDigest.begin(),dig);	
		mcnAggs.insert(mcnAggs.begin(),agg);
	}

	void updateMCNVOs(long agg, byte* digest)
	{
		byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		memcpy(dig, digest, DIGEST_LENGTH);
		mcnVOdigest.insert(mcnVOdigest.begin(),dig);	
		mcnVOaggs.insert(mcnVOaggs.begin(),agg);
	}
	
	void updateMCN(long agg, byte* digest, double* pLow, double* pHigh, int dimension)
	{
		byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		memcpy(dig, digest, DIGEST_LENGTH);
		mcnDigest.insert(mcnDigest.begin(),dig);	
		mcnAggs.insert(mcnAggs.begin(),agg);
		double* low=(double*)malloc(sizeof(double)*dimension);
		double* high=(double*)malloc(sizeof(double)*dimension);
		memcpy(low, pLow, dimension*sizeof(double));
		memcpy(high, pHigh, dimension*sizeof(double));
		mcnMBRs.insert(mcnMBRs.begin(), low);
		mcnMBRs.insert(mcnMBRs.begin(), high);
		m_dimension=dimension;
	}
		
	void updateMCNcounter(long level)
	{
		if (level==0)
			m_leafmcn++;
		else
			m_indexmcn++;
	}

	void updateMCNVOs(long agg, byte* digest, double* pLow, double* pHigh, int dimension)
	{
		byte* dig=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		memcpy(dig, digest, DIGEST_LENGTH);
		mcnVOdigest.insert(mcnVOdigest.begin(),dig);	
		mcnVOaggs.insert(mcnVOaggs.begin(),agg);
		double* low=(double*)malloc(sizeof(double)*dimension);
		double* high=(double*)malloc(sizeof(double)*dimension);
		memcpy(low, pLow, dimension*sizeof(double));
		memcpy(high, pHigh, dimension*sizeof(double));
		mcnVOMBRs.insert(mcnVOMBRs.begin(), low);
		mcnVOMBRs.insert(mcnVOMBRs.begin(), high);
		m_dimension=dimension;
	}

	void updateMCNaffectedNode(long nodeID)
	{
		bool found=false;
		for (int i=0; i<mcnAffectedNode.size(); i++)
		{
			if (mcnAffectedNode[i]==nodeID)
			{
				found=true;
				break;
			}
		}	
		if (!found)
			mcnAffectedNode.insert(mcnAffectedNode.begin(), nodeID);	
	}

	void computeAggAns()
	{
		long agg=0;
		cout<<"mcnAggs size "<<mcnAggs.size()<<endl;
		for (int i=0;i<mcnAggs.size();i++)
		{
			cout<<mcnAggs[i]<<endl;
			agg+=mcnAggs[i];
		}
		cout<<"Aggregation answer: "<<agg<<endl;
	}
	
	void calculateAggVOsize()
	{
		m_mcnSize=mcnDigest.size();
		m_mcnVOSize=mcnVOdigest.size();
		m_affectedNode=mcnAffectedNode.size();
		cout<<"mcn size "<<m_mcnSize<<endl;
		cout<<"mcn additional vo size "<<m_mcnVOSize<<endl;
		cout<<"affected nodes "<<m_affectedNode<<endl;
		//long size=4*(m_mcnSize+m_mcnVOSize)+DIGEST_LENGTH*(m_mcnSize+m_mcnVOSize)+2*m_dimension*8*(m_mcnSize+m_mcnVOSize);
	}
};

// example of a Strategy pattern.
// traverses the tree by level.
class MyQueryStrategy : public SpatialIndex::IQueryStrategy
{
private:
	queue<long> ids;

public:
	void getNextEntry(const IEntry& entry, long& nextEntry, bool& hasNext)
	{
		IShape* ps;
		entry.getShape(&ps);
		Region* pr = dynamic_cast<Region*>(ps);

		cout << pr->m_pLow[0] << " " << pr->m_pLow[1] << endl;
		cout << pr->m_pHigh[0] << " " << pr->m_pLow[1] << endl;
		cout << pr->m_pHigh[0] << " " << pr->m_pHigh[1] << endl;
		cout << pr->m_pLow[0] << " " << pr->m_pHigh[1] << endl;
		cout << pr->m_pLow[0] << " " << pr->m_pLow[1] << endl << endl << endl;
			// print node MBRs gnuplot style!

		delete ps;

		const INode* n = dynamic_cast<const INode*>(&entry);

		// traverse only index nodes at levels 2 and higher.
		if (n != 0 && n->getLevel() > 1)
		{
			for (unsigned long cChild = 0; cChild < n->getChildrenCount(); cChild++)
			{
				ids.push(n->getChildIdentifier(cChild));
			}
		}

		if (! ids.empty())
		{
			nextEntry = ids.front(); ids.pop();
			hasNext = true;
		}
		else
		{
			hasNext = false;
		}
	}
};

// example of a Strategy pattern.
// find the total indexed space managed by the index (the MBR of the root).
class MyQueryStrategy2 : public IQueryStrategy
{
public:
	Region m_indexedSpace;

public:
	void getNextEntry(const IEntry& entry, long& nextEntry, bool& hasNext)
	{
		// the first time we are called, entry points to the root.

		// stop after the root.
		hasNext = false;

		IShape* ps;
		entry.getShape(&ps);
		ps->getMBR(m_indexedSpace);
		delete ps;
	}
};

int main(int argc, char** argv)
{
	try
	{
		if (argc != 6)
		{
			cerr << "Usage: " << argv[0] << " query_file tree_file query_type [intersection | 10NN | selfjoin | contain | agg] dimension numberofquery." << endl;
			return -1;
		}

		int queryType = 0;

		if (strcmp(argv[3], "intersection") == 0) queryType = 0;
		else if (strcmp(argv[3], "10NN") == 0) queryType = 1;
		else if (strcmp(argv[3], "selfjoin") == 0) queryType = 2;
		else if (strcmp(argv[3], "contain") == 0) queryType=3;
		else if (strcmp(argv[3], "agg") == 0) queryType = 4;
		else
		{
			cerr << "Unknown query type." << endl;
			return -1;
		}

		ifstream fin(argv[1]);
		if (! fin)
		{
			cerr << "Cannot open query file " << argv[1] << "." << endl;
			return -1;
		}

		int dimension = atoi(argv[4]);
		int numberOfQuery = atoi(argv[5]);

		string baseName = argv[2];
		IStorageManager* diskfile = StorageManager::loadDiskStorageManager(baseName);
			// this will try to locate and open an already existing storage manager.

		StorageManager::IBuffer* file = StorageManager::createNewRandomEvictionsBuffer(*diskfile, 10, false);
			// applies a main memory random buffer on top of the persistent storage manager
			// (LRU buffer, etc can be created the same way).

		// If we need to open an existing tree stored in the storage manager, we only
		// have to specify the index identifier as follows
		ISpatialIndex* tree = RTree::loadRTree(*file, 1);

		unsigned long count = 0;
		unsigned long indexIO = 0;
		unsigned long leafIO = 0;
		unsigned long id, op;
		double* x=new double[dimension];
		double* y=new double[dimension];
		double* plow=new double[dimension];
		double* phigh=new double[dimension];
		Tools::ResourceUsage ru;
		unsigned long mcnSize = 0;
		unsigned long mcnVOSize = 0;
		unsigned long affectedNode = 0;
		unsigned long totalVOSize = 0;
		unsigned long leafmcn = 0;
		unsigned long indexmcn = 0;

		ru.start();

		while (fin)
		{
			fin >> op;
			for (int i=0; i<dimension; i++)
				fin >> x[i];
			for (int i=0; i<dimension; i++)
				fin >> y[i];
			if (! fin.good()) continue; // skip newlines, etc.

			if (op == QUERY)
			{
				for (int i=0;i<dimension;i++)
				{
					plow[i]=x[i];
					phigh[i]=y[i];
				}
				MyVisitor vis;

				if (queryType == 0)
				{
					Region r = Region(plow, phigh, dimension);
					tree->intersectsWithQuery(r, vis);
						// this will find all data that intersect with the query range.
				}
				else if (queryType == 1)
				{
					Point p = Point(plow, dimension);
					tree->nearestNeighborQuery(10, p, vis);
						// this will find the 10 nearest neighbors.
				}
				else if (queryType == 2)
				{
					Region r = Region(plow, phigh, dimension);
					tree->selfJoinQuery(r, vis);
				}
				else if (queryType == 3)
				{
					Region r = Region(plow, phigh, dimension);
					tree->containsWhatQuery(r, vis);
				}
				else if (queryType == 4)
				{
					Region r = Region(plow, phigh, dimension);
					tree->mcnQuery(r, vis);
					vis.calculateAggVOsize();
					mcnSize += vis.m_mcnSize;
					mcnVOSize += vis.m_mcnVOSize;
					affectedNode += vis.m_affectedNode;
					leafmcn += vis.m_leafmcn;
					indexmcn += vis.m_indexmcn;
				}

				indexIO += vis.m_indexIO;
				leafIO += vis.m_leafIO;
					// example of the Visitor pattern usage, for calculating how many nodes
					// were visited.
			}
			else
			{
				cerr << "This is not a query operation." << endl;
			}

			if ((count % 1000) == 0) cerr << count << endl;
			if (count==numberOfQuery) break;

			count++;
		}

		ru.stop();
		totalVOSize=(mcnSize+mcnVOSize)*(DIGEST_LENGTH+2*dimension*4+4);

		MyQueryStrategy2 qs;
		tree->queryStrategy(qs);

		cerr << "Indexed space: " << qs.m_indexedSpace << endl;
		cerr << "Operations: " << count << endl;
		cerr << *tree;
		cerr << "Index I/O: " << indexIO << endl;
		cerr << "Leaf I/O: " << leafIO << endl;
		cerr << "Buffer hits: " << file->getHits() << endl;
		cerr << "Seconds: " << ru.getTotalTime() << endl;
		cerr << "avg Index I/O: " << indexIO/count << endl;
		cerr << "avg Leaf I/O: " << leafIO/count << endl;
		cerr << "avg total I/O: " << (indexIO+leafIO)/count << endl;
		cerr << "avg mcnSize : " << mcnSize/count << endl;
		cerr << "avg leafmcn : " << leafmcn/count << endl;
		cerr << "avg indexmcn : " << indexmcn/count << endl;
		cerr << "avg mcnVOSize : "<<mcnVOSize/count << endl;
		cerr << "avg totalVOsize: "<<  totalVOSize/count << endl;
		cerr << "avg affectedNode: "<< affectedNode/count << endl;

		delete tree;
		delete file;
		delete diskfile;
			// delete the buffer first, then the storage manager
			// (otherwise the the buffer will fail writting the dirty entries).
	}
	catch (Tools::Exception& e)
	{
		cerr << "******ERROR******" << endl;
		std::string s = e.what();
		cerr << s << endl;
		return -1;
	}
	catch (...)
	{
		cerr << "******ERROR******" << endl;
		cerr << "other exception" << endl;
		return -1;
	}

	return 0;
}
