// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include "../spatialindex/SpatialIndexImpl.h"

#include "BTree.h"
#include "Node.h"
#include "Index.h"

#include "Leaf.h"

using namespace std;
using namespace SpatialIndex::BTree;

Leaf::~Leaf()
{
}

Leaf::Leaf(BTree* pTree, long id): Node(pTree, id, 0, pTree->m_leafCapacity)
{
	m_minFanout= (long)( ceil(((double)m_capacity)/2.0) );
}

NodePtr Leaf::chooseSubtree(const Point& key, unsigned long level, std::stack<long>& pathBuffer)
{
	// should make sure to relinquish other PoolPointer lists that might be pointing to the
	// same leaf.
	return NodePtr(this, &(m_pTree->m_leafPool));
}

NodePtr Leaf::findLeaf(const Point& key, long id, std::stack<long>& pathBuffer)
{
	for (unsigned long cChild = 0; cChild < m_children; cChild++)
	{
		// should make sure to relinquish other PoolPointer lists that might be pointing to the
		// same leaf.
		if (m_pIdentifier[cChild] == id && key == *(m_ptrKey[cChild])) return NodePtr(this, &(m_pTree->m_leafPool));
	}

	return NodePtr();
}

void Leaf::deleteData(long id, std::stack<long>& pathBuffer)
{
	unsigned long child;

	for (child = 0; child < m_children; child++)
	{
		if (m_pIdentifier[child] == id) break;
	}

	deleteEntry(child);
	m_pTree->writeNode(this);
}

void Leaf::showNode()
{
	int labelLen = m_pTree->m_stats.m_treeHeight-m_level;
	cout<<"Leaf Node "<<m_identifier<<" with "<<m_children<<" entries"<<endl;
	cout<<"Capcity "<<m_capacity<<" MinFanout "<<m_minFanout<<endl;	
	for (long cChild=0;cChild<m_children;cChild++)
	{
		cout<<" key="<<(*(m_ptrKey[cChild])).m_pCoords[0]<<" pid="<<m_pIdentifier[cChild]<<" agg="<<m_pAggValue[cChild];
		if (m_pLabel!=0)
		{
			printf(" label=");
			for (int j=0; j< labelLen; j++)
				printf("%d", m_pLabel[cChild][j]); 	
		}
		cout<<"	hashes=";
		char* d= printDigest(m_pDigest[cChild], DIGEST_LENGTH);		
		for (unsigned long i=0; i< DIGEST_LENGTH*2; i++)
			cout << d[i];
		delete[] d;
	}
	cout<<" next node="<<m_pFirstNode<<" ";
	cout<<endl;
	cout<<endl;
}
