// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include "../spatialindex/SpatialIndexImpl.h"

#include "BTree.h"
#include "Node.h"
#include "Index.h"

using std::stack;
using std::vector;
using namespace std;
using namespace SpatialIndex::BTree;

//
// Tools::IObject interface
//
Tools::IObject* Node::clone() throw (Tools::NotSupportedException)
{
	throw Tools::NotSupportedException("IObject::clone should never be called.");
}

//
// Tools::ISerializable interface
//
unsigned long Node::getByteArraySize()
{
	//int labelLen=m_pTree->m_stats.m_treeHeight-m_level;
	return
		(sizeof(long) +
		sizeof(long) +
		sizeof(long) +
		(m_children * ( sizeof(long)  + sizeof(long) + sizeof(long) + sizeof(int)*m_labelLen ) ) +
		//(m_children * ( sizeof(long)  + sizeof(long) + sizeof(long) ) ) +
		sizeof(int) + sizeof(long) +  sizeof(long) + sizeof(unsigned long) + m_totalDigestLength); 
}

unsigned long Node::loadFromByteArray(byte* const data)
{
	byte* ptr = data;
	long tmp;
	

	// skip the node type information, it is not needed.
	ptr += sizeof(long);

	memcpy(&m_level, ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&m_children, ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&m_totalDigestLength, ptr, sizeof(unsigned long));
	ptr += sizeof(unsigned long);
	
	memcpy(&(m_pFirstNode), ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&(m_pFirstAggValue), ptr, sizeof(long));
	ptr += sizeof(long);

	memcpy(&(m_labelLen), ptr, sizeof(int));
	ptr += sizeof(int);


	if (m_totalDigestLength >0 && m_level!=0)
	{
		m_pFirstDigest = new byte[DIGEST_LENGTH];
		memcpy(m_pFirstDigest, ptr, DIGEST_LENGTH);
		ptr += DIGEST_LENGTH;
	}
	else
		m_pFirstDigest=0;

	unsigned long cChild=0;
	for (cChild = 0; cChild < m_children; cChild++)
	{
		memcpy(&(m_pIdentifier[cChild]), ptr, sizeof(long));
		ptr += sizeof(long);

		memcpy(&(m_pAggValue[cChild]), ptr, sizeof(long));
		ptr += sizeof(long);

		m_ptrKey[cChild]=m_pTree->m_pointPool.acquire();
		*(m_ptrKey[cChild]) = m_pTree->m_infinitePoint;

		memcpy(&tmp, ptr, sizeof(long));
		ptr += sizeof(long);

		m_ptrKey[cChild]->m_pCoords[0]=tmp;

		m_pLabel[cChild]=new int[m_labelLen];
		for (int i=0;i<m_labelLen; i++)
		{
			memcpy(&(m_pLabel[cChild][i]), ptr, sizeof(int));
			ptr += sizeof(int);
		}


		if (m_totalDigestLength>0)
		{
			m_pDigest[cChild] = new byte[DIGEST_LENGTH];
			memcpy(m_pDigest[cChild], ptr, DIGEST_LENGTH);
			ptr += DIGEST_LENGTH;
		}
	}


	return getByteArraySize();
}

void Node::storeToByteArray(unsigned long& len, byte** data)
{

	m_labelLen=m_pTree->m_stats.m_treeHeight-m_level;
	len = getByteArraySize();

	*data = new byte[len];
	byte* ptr = *data;
		
	long tmp;

	long type;

	if (m_level == 0) type = PersistentLeaf;
	else type = PersistentIndex;

	memcpy(ptr, &type, sizeof(long));
	ptr += sizeof(long);

	memcpy(ptr, &m_level, sizeof(long));
	ptr += sizeof(long);

	memcpy(ptr, &m_children, sizeof(long));
	ptr += sizeof(long);
	
	memcpy(ptr, &m_totalDigestLength, sizeof(unsigned long));
	ptr += sizeof(unsigned long);

	memcpy(ptr, &(m_pFirstNode), sizeof(long));
	ptr += sizeof(long);

	memcpy(ptr, &(m_pFirstAggValue), sizeof(long));
	ptr += sizeof(long);
	
	memcpy(ptr, &(m_labelLen), sizeof(int));
	ptr += sizeof(int);

	if (m_totalDigestLength!= 0 && m_level!=0)
	{
		memcpy(ptr, m_pFirstDigest, DIGEST_LENGTH);
		ptr += DIGEST_LENGTH;
	}

	unsigned long cChild=0;
	for (cChild = 0; cChild < m_children; cChild++)
	{
		memcpy(ptr, &(m_pIdentifier[cChild]), sizeof(long));
		ptr += sizeof(long);

		memcpy(ptr, &(m_pAggValue[cChild]), sizeof(long));
		ptr += sizeof(long);

		tmp=(long)(m_ptrKey[cChild]->m_pCoords[0]);
		memcpy(ptr, &tmp, sizeof(long));
		ptr += sizeof(long);
		
		if (m_pLabel[cChild]==0)
		{	m_pLabel[cChild]=new int[m_labelLen];
			for (int i=0;i<m_labelLen;i++)
				m_pLabel[cChild][i]=0;
		}

		for (int i=0;i<m_labelLen; i++)
		{
			memcpy(ptr, &(m_pLabel[cChild][i]), sizeof(int));
			ptr += sizeof(int);
		}
		
		if (m_totalDigestLength!= 0)
		{
			memcpy(ptr, m_pDigest[cChild], DIGEST_LENGTH);
			ptr += DIGEST_LENGTH;
		}
	}
	if (len!=(ptr-*data))
	{
		cout<<"node id "<<m_identifier<<endl;
		cout<<len<<" "<<(ptr-*data)<<" "<<m_totalDigestLength<<" "<<m_children<<endl;
	}
	assert(len == (ptr - *data) );
}

//
// SpatialIndex::IEntry interface
//
long Node::getIdentifier() const
{
	return m_identifier;
}

void Node::getShape(IShape** out) const throw (std::exception)
{
	//*out = m_pTree->readNode(m_pFirstNode).get();

}

//
// SpatialIndex::INode interface
//
unsigned long Node::getChildrenCount() const
{
	return m_children;
}

long Node::getChildIdentifier(unsigned long index) const throw (Tools::IndexOutOfBoundsException)
{
	if (index < -1 || index >= (m_children)) throw Tools::IndexOutOfBoundsException(index);
	
	if (index>=0)
		return m_pIdentifier[index];
	return m_pFirstNode;
}

void Node::getChildShape(unsigned long index, IShape** out) const throw (Tools::IndexOutOfBoundsException, std::exception)
{
	if (index < 0 || index >= m_children) throw Tools::IndexOutOfBoundsException(index);

	*out = new Point(*(m_ptrKey[index]));
}

unsigned long Node::getLevel() const
{
	return m_level;
}

bool Node::isLeaf() const
{
	return (m_level == 0);
}

bool Node::isIndex() const
{
	return (m_level != 0);
}

//
// Internal
//

Node::Node() :
	m_pTree(0),
	m_level(0),
	m_identifier(-1),
	m_children(0),
	m_capacity(0),
	m_ptrKey(0),
	m_pIdentifier(0),
	m_pFirstNode(-1),
	m_pDigest(0),
	m_pFirstDigest(0),
	m_totalDigestLength(0),
	m_pFirstAggValue(0),
	m_pAggValue(0),
	m_pLabel(0)
{
}

Node::Node(BTree* pTree, long id, unsigned long level, unsigned long capacity) :
	m_pTree(pTree),
	m_level(level),
	m_identifier(id),
	m_children(0),
	m_capacity(capacity),
	m_ptrKey(0),
	m_pIdentifier(0),
	m_pFirstNode(-1),
	m_pDigest(0),
	m_pFirstDigest(0),
	m_totalDigestLength(0),
	m_pFirstAggValue(0),
	m_pAggValue(0),
	m_pLabel(0)
{

	try
	{
		m_ptrKey = new PointPtr[m_capacity];
		m_pIdentifier = new long[m_capacity];
		m_pAggValue = new long[m_capacity];
		m_pDigest = new byte*[m_capacity];
		memset(m_pDigest,0,sizeof(byte*)*m_capacity);		
		m_pLabel= new int*[m_capacity];
		memset(m_pLabel,0,sizeof(int*)*m_capacity);		

		//m_minFanout = (m_capacity+1)>>1;
	}
	catch (...)
	{
		delete[] m_ptrKey;
		delete[] m_pIdentifier;
		delete[] m_pAggValue;
		throw;
	}
}

Node::~Node()
{
	if (m_pDigest != 0)
	{
		for (unsigned long cChild = 0; cChild < m_children; cChild++)
		{
			if (m_pDigest[cChild] != 0) delete[] m_pDigest[cChild];
		}

		delete[] m_pDigest;
	}
	if (m_pFirstDigest!=0)
		delete[] m_pFirstDigest;

	delete[] m_ptrKey;
	delete[] m_pIdentifier;
	delete[] m_pAggValue;
}

Node& Node::operator=(const Node& n)
{
	throw Tools::IllegalStateException("operator =: This should never be called.");
}


//deleteEntry is never called, deletetion is done through remove entry
void Node::deleteEntry(unsigned long index)
{
	assert(index >= 0 && index < m_children);

	PointPtr ptrK = m_ptrKey[index];
	m_totalDigestLength -= DIGEST_LENGTH;
	if (m_pDigest[index] != 0) delete[] m_pDigest[index];


	if (m_children > 1 && index != m_children - 1)
	{
		m_ptrKey[index] = m_ptrKey[index+1];
		m_pIdentifier[index] = m_pIdentifier[index +1];
		m_pAggValue[index] = m_pAggValue[index+1];
		m_pDigest[index] = m_pDigest[index+1];
	}

	m_children--;

	// WARNING: index has now changed. Do not use it below here.

	if (m_children == 0)
	{
	}
	//	m_nodeMBR.m_pLow[cDim] = std::numeric_limits<double>::max();
	//	m_nodeMBR.m_pLow[cDim] = std::min(m_nodeMBR.m_pLow[cDim], m_ptrMBR[cChild]->m_pLow[cDim]);
}


bool Node::isFull()
{
	bool full=false;
	if (m_children==m_capacity)
		full=true;
	return full;
}

bool Node::isFewest()
{
	bool fewest=false;
	if (m_level==m_pTree->m_stats.m_treeHeight-1)
	{
		if (m_children==1)
			fewest=true;
	}
	else if (m_children<=m_minFanout)
		fewest=true;

	return fewest;
}


long Node::getSlot(double searchKey)
{
	long slot, entries;
	
	entries=m_children;
	if (entries==0)
	 	slot=-1;
	else	slot=findKey(searchKey, 0, entries);
	return slot;
}

long Node::findKey(double searchKey, long lo, long hi)
{
	long mid;
	long findslot;
	
	if (hi==lo)
		findslot=bestMatch(searchKey, lo);
	else
	{
		mid = (lo + hi) >> 1;
		switch (findslot = bestMatch(searchKey, mid))
		{
			case -2:
				findslot = findKey(searchKey, lo, mid-1);
				break;
			case -3: 
				findslot= findKey(searchKey, mid+1, hi);
				break;
			default:
				break;
		}
	}
	return findslot;
}

long Node::bestMatch(double searchKey,long slot)
{
	double diff, comp;
	long findslot;

	diff=this->m_pTree->compareKeys(searchKey, m_ptrKey[slot]->m_pCoords[0]);
   	if (diff < 0.0)
	{
		if (slot==0)
			findslot=-1;
		else if ((comp=this->m_pTree->compareKeys(searchKey, m_ptrKey[slot-1]->m_pCoords[0]))>=0.0)
			findslot=slot-1;
		else
			findslot= -2;
	}
	else
	{
		if (slot==m_children-1)
			findslot=slot;
		else if ((comp=this->m_pTree->compareKeys(searchKey, m_ptrKey[slot+1]->m_pCoords[0]))<0.0)
			findslot=slot;
		else if (comp==0.0)
			findslot=slot+1;
		else
			findslot=-3;
	}
	return findslot;
}

void Node::insertEntry(long slot, Point& key, long id, long agg, NodePtr sibling, NodePtr downPtr)
{
	long split, i, j, k, x, y;
	if (sibling.get()==0)
	{
		placeEntry(slot+1, key, id, agg, downPtr);
	}
	else
	{
		if (this->isIndex())
			i=1;
		else
			i=0;
		if (i==1)
			split=(m_capacity-m_minFanout)-1;
		else
			split=m_minFanout-1;
	        if (slot!=split)
			j=1;
		else
			j=0;
		if (slot>=split)
			k=1;
		else
			k=0;
		//cout<<"insertEntry "<<"i "<<i<<"j "<<j<<"k "<<k<<"slot "<<slot<<"split "<<split<<endl;
		for (x=split+k+i*j, y=0; x<m_capacity; x++, y++)
		{
			sibling->m_ptrKey[y]=m_pTree->m_pointPool.acquire();
			*(sibling->m_ptrKey[y])=*(this->m_ptrKey[x]);
			sibling->m_pIdentifier[y]=this->m_pIdentifier[x];
			sibling->m_pAggValue[y]=this->m_pAggValue[x];
			sibling->m_pDigest[y] = this->m_pDigest[x];
			this->m_children--;
			sibling->m_children++;
			this->m_totalDigestLength-=DIGEST_LENGTH;
			sibling->m_totalDigestLength+=DIGEST_LENGTH;
		}	
	       if (i==1)
		{
			if (j==1)
			{
				sibling->m_pFirstNode=this->m_pIdentifier[split+k];
				sibling->m_pFirstAggValue=this->m_pAggValue[split+k];
				sibling->m_pFirstDigest=this->m_pDigest[split+k];
				this->m_children--; //as soon the new entry will be inserted and replace this entry at split+k;
				sibling->m_totalDigestLength+=DIGEST_LENGTH;
				this->m_totalDigestLength-=DIGEST_LENGTH;
			}
			else
			{
				sibling->m_pFirstNode=downPtr->m_identifier;
				sibling->m_pFirstAggValue=downPtr->computeNodeAggValue();
				byte* digest=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));	
				sibling->computeNodeDigests(downPtr, digest);
				sibling->m_pFirstDigest=digest;
				sibling->m_totalDigestLength+=DIGEST_LENGTH;
			}
		}		
		if (j==1)
		{
			Point tmpKey=*(m_ptrKey[split+k]);
			if (this->isLeaf())
			{	m_pTree->m_tmpKey=*(m_ptrKey[split+k]);
				//cout<<"update m_tmpKey "<<m_pTree->m_tmpKey.m_pCoords[0]<<endl;
			}
			if (k==1)
			{
				sibling->placeEntry(slot-split-i, key, id, agg, downPtr);
			}
			else
				this->placeEntry(slot+1, key, id, agg, downPtr);
			if (this->isIndex())
			{
				m_pTree->m_tmpKey=tmpKey;
				//cout<<"update m_tmpKey "<<m_pTree->m_tmpKey.m_pCoords[0]<<endl;
			}
		}
		else
		{
			m_pTree->m_tmpKey=key;
			if (i!=1)
				sibling->placeEntry(0, key, id, agg, downPtr);
		}
	}
}

void Node::placeEntry(long slot, Point& key, long id, long agg, NodePtr downPtr)
{
	long i,x;
	if (m_children>0)
	{
		this->m_ptrKey[m_children]=m_pTree->m_pointPool.acquire();
		for (x=this->m_children-1; x>=slot; x--)
		{
			this->m_pIdentifier[x+1]=this->m_pIdentifier[x];
			this->m_pAggValue[x+1]=this->m_pAggValue[x];
			*(this->m_ptrKey[x+1])=*(this->m_ptrKey[x]);		
			m_pDigest[x+1] = m_pDigest[x];
		}
	}
	if (slot==0)
		this->m_ptrKey[slot]=m_pTree->m_pointPool.acquire();
	if (this->isIndex())
	{
		//compute hashes here or pass hashes in
		byte* digest=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));	
		this->computeNodeDigests(downPtr,digest);
		if (slot==-1)
		{
			m_pFirstNode=downPtr->m_identifier;
			m_pFirstAggValue=downPtr->computeNodeAggValue();
			m_pFirstDigest=digest;
			m_totalDigestLength+=DIGEST_LENGTH;
		}
		else
		{
			this->m_pIdentifier[slot]=downPtr->m_identifier;
			this->m_pAggValue[slot]=downPtr->computeNodeAggValue();
			*(this->m_ptrKey[slot])=m_pTree->m_tmpKey;
			this->m_pDigest[slot]=digest;
			this->m_totalDigestLength+=DIGEST_LENGTH;
		}
	}			
	else
	{
		this->m_pIdentifier[slot]=id;
		this->m_pAggValue[slot]=agg;
		//this->m_ptrKey[slot]=m_pTree->m_pointPool.acquire();
		*(this->m_ptrKey[slot])=key;

		//compute hashes here or pass hashes in
		byte* output=(byte*)malloc(DIGEST_LENGTH*sizeof(byte));
		long keyvalue=(long)key.m_pCoords[0];
		unsigned char* input=(unsigned char*)malloc(2*sizeof(long));
		memcpy(input, &keyvalue, sizeof(long));
		memcpy(input+sizeof(long), &agg, sizeof(long));
		
		m_pTree->m_c0=clock();

		sha1(input, output, 2*sizeof(long));			
		
		m_pTree->m_c1=clock();
		m_pTree->m_cputime +=(double)(m_pTree->m_c1-m_pTree->m_c0)/(double)(CLOCKS_PER_SEC);
		m_pTree->m_hashComputed++;
	
		this->m_pDigest[slot]=output;
		this->m_totalDigestLength+=DIGEST_LENGTH;
	}
	this->m_children++;
}

void Node::computeNodeDigests(NodePtr node, byte* digest)
{
	//cout<<"compute digest for node "<<node->m_identifier<<endl;
	
	if (node->m_totalDigestLength==0)
		digest=0;
	else
	{
		byte* input=(byte*)malloc(node->m_totalDigestLength*sizeof(byte)+(node->m_children+1)*sizeof(long));
		byte* tmp=input;
		if(node->isIndex())
		{	
			memcpy(tmp, node->m_pFirstDigest, DIGEST_LENGTH);
			tmp+=DIGEST_LENGTH;
			memcpy(tmp, &(node->m_pFirstAggValue), sizeof(long));
			tmp+=sizeof(long);
		}
		for (unsigned long i=0; i<node->m_children;i++)
		{
			memcpy(tmp, node->m_pDigest[i],DIGEST_LENGTH);
			tmp+=DIGEST_LENGTH;
			memcpy(tmp, &(node->m_pAggValue[i]), sizeof(long));
			tmp+=sizeof(long);
		}
		m_pTree->m_c0=clock();
		if (node->isIndex())
			sha1(input, digest, node->m_totalDigestLength+(node->m_children+1)*sizeof(long));			
		else
			sha1(input, digest, node->m_totalDigestLength+(node->m_children)*sizeof(long));			
		m_pTree->m_c1=clock();
		m_pTree->m_cputime +=(double)(m_pTree->m_c1-m_pTree->m_c0)/(double)(CLOCKS_PER_SEC);
		m_pTree->m_hashComputed++;
	}

}

long Node::computeNodeAggValue()
{
	long agg=0;
	for (long x=0; x<m_children; x++)
		agg+=m_pAggValue[x];
	return agg;
}

void Node::removeEntry(long slot)
{
	long x;
	if (this->isIndex())
		m_pTree->deleteNode( (m_pTree->readNode(this->m_pIdentifier[slot])).get() );
	m_totalDigestLength -= DIGEST_LENGTH;
	if (m_pDigest[slot]!=0)
		delete[] m_pDigest[slot];
	for (x=slot; x<m_children-1; x++)
	{
		m_pIdentifier[x]=m_pIdentifier[x+1];
		m_pAggValue[x]=m_pAggValue[x+1];
		*(m_ptrKey[x])=*(m_ptrKey[x+1]);
		m_pDigest[x] = m_pDigest[x+1];
	}
	this->m_children--;
	return;
}

void Node::constructLabel(long pos, int treeHeight)
{
	long i,slot;
	int labelLen=treeHeight-m_level;
	int x;
	m_pLabel=(int**)malloc(m_children*sizeof(int*));
	memset(m_pLabel, 0, sizeof(int*)*m_children);
	for (i=0;i<m_children;i++)
	{
		//if (m_pLabel[i]==0)
		m_pLabel[i]=(int*)malloc(sizeof(int)*labelLen);
		for (int k=0;k<labelLen;k++)
			m_pLabel[i][k]=0;
		int j=labelLen-1;
		long vpos=pos+i;	
		while (vpos>=m_capacity)
		{
			x=(int)(vpos-(long)(floor((double)vpos/(double)m_capacity)*(double)m_capacity));		
			vpos=(long)floor((double)vpos/(double)m_capacity);
			//m_pLabel[i][j]=x+('0'-0);		
			m_pLabel[i][j]=x;
			j--;
		}
		x=vpos%m_capacity;
		//m_pLabel[i][j]=x+('0'-0);		
		m_pLabel[i][j]=x;
	}
}
