// Spatial Index Library
//
// Copyright (C) 2003 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <SpatialIndex.h>

using namespace SpatialIndex;

NwayNode::NwayNode() :
	m_level(0),
	m_identifier(-1),
	m_children(0),
	m_fanout(0),
	m_label(0),
	m_pDigest(0),
	m_pChild(0),
	m_pChildID(0),
	m_aggvalue(0),
	m_pTree(0)
{
}

NwayNode::NwayNode(NwayTree* tree) :
	m_level(0),
	m_identifier(-1),
	m_children(0),
	m_fanout(0),
	m_label(0),
	m_pDigest(0),
	m_pChild(0),
	m_pChildID(0),
	m_pTree(tree)	
{
	m_fanout=m_pTree->m_fanout;
	m_label=new int*[m_fanout];
	memset(m_label, 0, sizeof(int*)*m_fanout);
	m_pChild= new NwayNode*[m_fanout];
	memset(m_pChild, 0, sizeof(NwayNode*)*m_fanout);
	m_pChildID = new long[m_fanout];
	memset(m_pChildID, 0, sizeof(long)*m_fanout);
	m_pDigest= new byte*[m_fanout];
	memset(m_pDigest, 0, sizeof(byte*)*m_fanout);
	m_aggvalue=new long[m_fanout];
	memset(m_aggvalue, 0, sizeof(long)*m_fanout);

}

NwayNode::NwayNode(NwayTree* tree, byte* const data):
	m_level(0),
	m_identifier(-1),
	m_children(0),
	m_fanout(0),
	m_label(0),
	m_pDigest(0),
	m_pChild(0),
	m_pChildID(0),
	m_pTree(tree)	

{
	m_fanout=m_pTree->m_fanout;
	m_label=new int*[m_fanout];
	memset(m_label, 0, sizeof(int*)*m_fanout);
	m_pChild= new NwayNode*[m_fanout];
	memset(m_pChild, 0, sizeof(NwayNode*)*m_fanout);
	m_pChildID= new long[m_fanout];
	memset(m_pChildID, 0, sizeof(long)*m_fanout);
	m_pDigest= new byte*[m_fanout];
	memset(m_pDigest, 0, sizeof(byte*)*m_fanout);
	m_aggvalue=new long[m_fanout];
	memset(m_aggvalue, 0, sizeof(long)*m_fanout);

	loadFromByteArray(data);
}

NwayNode::NwayNode(NwayTree* tree, long level, long identifier, long children, long* aggs, NwayNode** childs, long* childsID, byte** digests):
	m_level(level),
	m_identifier(identifier),
	m_children(children),
	m_fanout(0),
	m_label(0),
	m_pDigest(0),
	m_pChild(0),
	m_pChildID(0),
	m_pTree(tree)	

{
	m_fanout=m_pTree->m_fanout;
	long i;
	long dividefactor=(long)pow((double)m_fanout, (double)level);
	m_label=new int*[m_fanout];
	memset(m_label, 0, sizeof(int*)*m_fanout);
	if (level==0)
		m_aggvalue=new long[m_fanout];
	else
		m_aggvalue=0;
	m_pChildID=new long[m_fanout];
	memset(m_pChildID, 0, sizeof(long)*m_fanout);
	m_pChild= new NwayNode*[m_fanout];
	memset(m_pChild, 0, sizeof(NwayNode*)*m_fanout);
	m_pDigest= new byte*[m_fanout];
	memset(m_pDigest, 0, sizeof(byte*)*m_fanout);
	for (i=0;i<m_children;i++)
	{
		//m_label[i]= new byte[m_pTree->m_treeHeight-m_level];
		m_label[i]=0;
		if (level==0)
			m_aggvalue[i]= aggs[i];
		if (childs!=0)
			m_pChild[i]=childs[i];
		if (childsID!=0)
			m_pChildID[i]=childsID[i];
		m_pDigest[i]=new byte[DIGEST_LENGTH];
		if (digests!=0)
		{
			memcpy(m_pDigest[i], digests[i], DIGEST_LENGTH);
			delete[] digests[i];
		}	
		else
		{
			if (m_level==0)
			{
				byte* input=(unsigned char*)&(m_aggvalue[i]);
				sha1(input, m_pDigest[i],sizeof(long));	
			}
		}
	}
	/*delete[] aggs;
	if (childs!=0)
		delete[] childs;
	if (childsID!=0)
		delete[] childsID;	
	if (digests!=0)
		delete[] digests;*/
}
	
NwayNode::~NwayNode()
{
	long i=0;
	for (i=0;i<m_children;i++)
	{
		if (m_pDigest[i]!=0)
			delete m_pDigest[i];
		if (m_pChild!=0 && m_pChild[i]!=0)
			delete m_pChild[i];
	}
	if (m_pDigest!=0)
		delete[] m_pDigest;
	if (m_label!=0)
		delete[] m_label;
	if (m_pChild!=0)
		delete[] m_pChild;
	if (m_pChildID!=0)
		delete[] m_pChildID;
	if (m_aggvalue!=0)
		delete[] m_aggvalue;
}

//
// IObject interface
//
NwayNode* NwayNode::clone() throw (Tools::NotSupportedException)
{
	return new NwayNode();
}

//
// ISerializable interface
//
unsigned long NwayNode::getByteArraySize()
{
	int labelLen= m_pTree->m_treeHeight-m_level;
	unsigned long nodeLen= sizeof(long) + sizeof(long) + sizeof(long) + m_children*(labelLen*sizeof(int) + sizeof(long)+ DIGEST_LENGTH);
	if (m_level==0)
		nodeLen+=m_children*(sizeof(long));
	return nodeLen;
}

unsigned long NwayNode::loadFromByteArray(byte* const data)
{
	byte* ptr = data;
	
	memcpy(&m_level, ptr, sizeof(long));
	ptr+=sizeof(long);
	
	memcpy(&m_identifier, ptr, sizeof(long));
	ptr+=sizeof(long);
		

	memcpy(&m_children, ptr, sizeof(long));
	ptr+=sizeof(long);

	for (long i=0; i<m_children; i++)
	{
		m_label[i]=new int[m_pTree->m_treeHeight-m_level];
		memcpy(m_label[i], ptr, (m_pTree->m_treeHeight-m_level)*sizeof(int));
		ptr+=(m_pTree->m_treeHeight-m_level)*sizeof(int);

		memcpy(&(m_pChildID[i]), ptr, sizeof(long));
		ptr+=sizeof(long);
		
		if (m_level==0)
		{
			memcpy(&(m_aggvalue[i]), ptr, sizeof(long));
			ptr+=sizeof(long);
		}		

		m_pDigest[i]=new byte[DIGEST_LENGTH];	
		memcpy(m_pDigest[i], ptr, DIGEST_LENGTH);
		ptr+=DIGEST_LENGTH;
	
	}
	return getByteArraySize();
}

void NwayNode::storeToByteArray(unsigned long& len, byte** data)
{
	len = getByteArraySize();
	*data = new byte[len];
	byte* ptr = *data;
	
	memcpy(ptr, &m_level, sizeof(long));
	ptr+=sizeof(long);
	
	memcpy(ptr, &m_identifier, sizeof(long));
	ptr+=sizeof(long);

	memcpy(ptr, &m_children, sizeof(long));
	ptr+=sizeof(long);

	for (long i=0; i<m_children; i++)
	{
		memcpy(ptr, m_label[i], (m_pTree->m_treeHeight-m_level)*sizeof(int));
		ptr+=(m_pTree->m_treeHeight-m_level)*sizeof(int);

		memcpy(ptr, &(m_pChildID[i]),sizeof(long));
		ptr+=sizeof(long);
		
		if (m_level==0)
		{
			memcpy(ptr, &(m_aggvalue[i]),sizeof(long));
			ptr+=sizeof(long);
		}

		memcpy(ptr, m_pDigest[i], DIGEST_LENGTH);
		ptr+=DIGEST_LENGTH;
	}
	assert(len==(ptr-*data));
}

void NwayNode::computeNodeDigests(byte* digest)
{
	if (m_children==0 || m_pDigest==0)
		digest=0;
	else if (m_children==1)
		memcpy(digest, m_pDigest[0], DIGEST_LENGTH);	
	else
	{
		byte* input=(byte*)malloc(m_children*DIGEST_LENGTH);
		byte* tmp=input;
		for (long i=0; i<m_children;i++)
		{
			memcpy(tmp, m_pDigest[i],DIGEST_LENGTH);
			tmp+=DIGEST_LENGTH;
		}
		m_pTree->m_c0=clock();
		sha1(input, digest, m_children*DIGEST_LENGTH);			
		m_pTree->m_c1=clock();
		m_pTree->m_cputime+=(double)(m_pTree->m_c1-m_pTree->m_c0)/(double)(CLOCKS_PER_SEC);
		m_pTree->m_hashComputed++;
		delete[] input;
	}
}

void NwayNode::constructLabel(long pos)
{
	long i,slot;
	int labelLen=m_pTree->m_treeHeight-m_level;
	int x;
	for (i=0;i<m_children;i++)
	{
		if (m_label[i]==0)
			m_label[i]=new int[labelLen];
		memset(m_label[i], 0, sizeof(int)*labelLen);
		int j=labelLen-1;
		long vpos=pos+i;	
		while (vpos>=m_fanout)
		{
			x=(int)(vpos-(long)(floor((double)vpos/(double)m_fanout)*(double)m_fanout));		
			vpos=(long)floor((double)vpos/(double)m_fanout);
			//m_label[i][j]=x+('0'-0);		
			m_label[i][j]=x;
			j--;
		}
		x=vpos%m_fanout;
		//m_label[i][j]=x+('0'-0);		
		m_label[i][j]=x;
	}
}

long NwayNode::getSlot(int* indexLabel)
{
	long i,j,slot;
	int labelLen=m_pTree->m_treeHeight-m_level;
	bool match;
	for (i=0;i<m_children;i++)
	{
		match=true;
		for (j=0;j<labelLen;j++)
		{
			if (m_label[i][j]!=indexLabel[j])
			{
				match=false;
				break;
			}
		}
		if (match==true)
			break;
	}
	slot=i;
	return slot;
}

void NwayNode::showNode()
{
	long i;
	int j;
	if (m_level==0)
		printf("Leaf node %d Level %d\n", m_identifier, m_level);
	else
		printf("Index node %d Level %d\n", m_identifier, m_level);
	int labelLen=m_pTree->m_treeHeight-m_level;
	for (i=0; i<m_children; i++)
	{
		//printf("key %d child %d ", m_label[i], m_pChildID[i]);
		printf("label ");
		for (j=0; j< labelLen; j++)
			printf("%d", m_label[i][j]); 	
		printf(" ");
		if (m_level==0)
			printf("aggval %d ", m_aggvalue[i]);
		for (j=0; j<DIGEST_LENGTH; j++)
			printf("%02x",m_pDigest[i][j]); 
		printf("\n");
	}
}

std::ostream& SpatialIndex::operator<<(std::ostream& os, const NwayNode& pt)
{

	return os;
}
