// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double m_xmin, m_xmax;

	Point(double x1, double x2)
	{
		m_xmin = (x1 < x2) ? x1 : x2;
		m_xmax = (x1 > x2) ? x1 : x2;
	}
};

int main(int argc, char** argv)
{
	if (argc != 3)
	{
		cerr << "Usage: " << argv[0] << " number_of_data time_instants." << endl;
		return -1;
	}

	unsigned long simulationLength = atol(argv[2]);
	unsigned long numberOfObjects = atol(argv[1]);
	map<unsigned long, Point> data;
	Tools::Random rnd;

	for (unsigned long i = 0; i < numberOfObjects; i++)
	{
		double x = rnd.nextUniformDouble();
		x=ceil(x*RANGE);
		double y = rnd.nextUniformDouble();
		y=ceil(y*RANGE);

		Point r = Point(x, y);

		data.insert(pair<unsigned long, Point>(i, r));

		cout << INSERT << " " << i << " " << r.m_xmin <<" "<< r.m_xmax << endl;
	}

	if (simulationLength == 0)
	{
		for (unsigned long i = 0; i < 10; i++)
		{
			unsigned long id = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects));
			map<unsigned long, Point>::iterator itMap = data.find(id);
			assert(itMap != data.end());
			cout << QUERY << " " << id <<" " << (*itMap).second.m_xmin << " "<<(*itMap).second.m_xmax << endl;
		}
	}

	unsigned long A = static_cast<unsigned long>(floor(static_cast<double>(numberOfObjects) * 0.05));

	set<unsigned long> examined;
	for (unsigned long T = 1; T <= simulationLength; T++)
	{
		cerr << (simulationLength + 1 - T) << endl;
		for (unsigned long a = 0; a < A; a++)
		{
			// find an id that is not yet examined.
			unsigned long id = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects));
			set<unsigned long>::iterator itSet = examined.find(id);

			while (itSet != examined.end())
			{
				id = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects));
				itSet = examined.find(id);
			}
			examined.insert(id);

			map<unsigned long, Point>::iterator itMap = data.find(id);
			assert(itMap != data.end());

			cout << DELETE << " " << id << " " << (*itMap).second.m_xmin << " "<< (*itMap).second.m_xmax << endl;
			data.erase(itMap);
		}

		unsigned long id = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects));
		set<unsigned long>::iterator itSet = examined.find(id);
		while (itSet != examined.end())
		{
			id = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects));
			itSet = examined.find(id);
		}
		map<unsigned long, Point>::iterator itMap = data.find(id);
		assert(itMap != data.end());
		cout << QUERY << " "<<id<<" " << (*itMap).second.m_xmin <<" "<< (*itMap).second.m_xmax << endl;
	}

	return 0;
}
