#include <SpatialIndex.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

using namespace SpatialIndex;
using namespace std;

//constructor
RBTree::RBTree()
{
	nil = new RBNode(0,BLACK);
	m_root = NULL;
	m_rootID = 0;
	m_id = 0;
	m_cputime=0.0;
	m_hashComputed=0;
	m_treeDigest=NULL;
}

RBTree::~RBTree()
{
	if (m_root!=NULL)
		delete m_root;
	if (m_treeDigest!=NULL)
		delete m_treeDigest;
}

unsigned long RBTree::getByteArraySize()
{
	unsigned long len=sizeof(int) + sizeof(int)+DIGEST_LENGTH;
	return len;
}

unsigned long RBTree::loadFromByteArray(byte* const data)
{
	byte* ptr = data;

	memcpy(&m_rootID, ptr, sizeof(int));
	ptr+=sizeof(int);

	memcpy(&m_id, ptr, sizeof(int));
	ptr+=sizeof(int);

	m_treeDigest= new byte[DIGEST_LENGTH];
	memcpy(m_treeDigest, ptr, DIGEST_LENGTH);
	ptr+=DIGEST_LENGTH;
	
	return getByteArraySize();
}

void RBTree::storeToByteArray(unsigned long& len, byte** data)
{
	len = getByteArraySize();
	*data = new byte[len];
	byte* ptr = *data;
	
	memcpy(ptr, &m_rootID, sizeof(int));
	ptr+=sizeof(int);
	
	memcpy(ptr, &m_id, sizeof(int));
	ptr+=sizeof(int);

	memcpy(ptr, m_treeDigest, DIGEST_LENGTH);
	ptr+=DIGEST_LENGTH;
	
	assert(len==(ptr-*data));
}

void RBTree::writeNode(RBNode* node, byte** buffer, unsigned long& dataLength)
{
	node->storeToByteArray(dataLength, buffer);	
}

void RBTree::writeTree(byte** buffer, unsigned long& dataLength)
{
	std::vector<RBNode*> nodes;
	std::vector<byte*> datas;
	std::vector<unsigned long> datalen;
	byte* tmp;
	unsigned long tmpLen=0;
	dataLength=0;
	if (m_root==0)
		return;
	this->storeToByteArray(tmpLen, &tmp);
	datas.push_back(tmp);
	datalen.push_back(tmpLen);	
	dataLength+=tmpLen;	
	nodes.push_back(m_root);
	while(nodes.size()>0)
	{
		RBNode* node=(RBNode*)nodes.front();
		nodes.erase(nodes.begin());
		writeNode(node, &tmp, tmpLen);
		datas.push_back(tmp);
		datalen.push_back(tmpLen);
		dataLength+=tmpLen;
		if (node->m_leftID!=-1)
			nodes.push_back(node->left);
		if (node->m_rightID!=-1)
			nodes.push_back(node->right);
	}
	*buffer= new byte[dataLength];
	byte* ptr = *buffer;
	long size=datas.size();
	for (long i=0; i<size; i++)
	{
		tmp=(byte*)datas.front();
		datas.erase(datas.begin());
		tmpLen=(long)datalen.front();
		datalen.erase(datalen.begin());
		memcpy(ptr, tmp, tmpLen);
		ptr+=tmpLen;
		delete[] tmp;
	}
}

RBNode* RBTree::readNode(byte*& buffer)
{
	RBNode* n=new RBNode(this);
	unsigned long len=n->loadFromByteArray(buffer);
	buffer+=len; 
	return n;
}

void RBTree::readTree(byte* buffer)
{
	byte* tmp=buffer;
	std::vector<RBNode*> nodes;
	unsigned long len=this->loadFromByteArray(buffer);
	buffer+=len;
	long i;
	for (i=0; i<m_id;i++)
	{
		RBNode* node = readNode(buffer);
		nodes.push_back(node);
	}	
	long size=nodes.size();	
	for (i=0; i<size; i++)
	{
		RBNode* node=(RBNode*)nodes.front();
		nodes.erase(nodes.begin());
		long set=0;
		for (long j=0; j<nodes.size();j++)
		{
			RBNode* tn = (RBNode*)nodes[j];
			for (long k=1; k<=node->m_child; k++)
			{
				if (k==1)
				{
					if (tn->m_id==node->m_leftID)
					{	
						node->left=tn;
						set++;
						break;		
					}	
				}
				if (k==2)
				{
					if (tn->m_id==node->m_rightID)
					{	
						node->right=tn;
						set++;
						break;		
					}	
				}
			}
			if (set==node->m_child)
				break;
		}
		if (node->m_id==m_rootID)
			m_root=node;
	}
	buffer=tmp;
}

void RBTree::storeToFile(char* file)
{
	int treeFile;
	int cMode = O_CREAT | O_RDWR | O_TRUNC;
	treeFile = open(file, cMode, 0644);
	ssize_t byteswritten;	
	byte* buffer;
	unsigned long length;
	writeTree(&buffer, length);
	//std::cout<<"tree length "<<length<<std::endl;
	byteswritten=write(treeFile, buffer, length);
	//fsync(treeFile);
}

void RBTree::loadFromFile(char* file)
{
	int treeFile;
	int cMode = O_RDWR;
	treeFile = open(file, cMode, 0644);
	ssize_t bytesread;
	unsigned long length;
	off_t seeke = lseek(treeFile, 0, SEEK_END);
	lseek(treeFile, -1*seeke, SEEK_END);
	length = (unsigned long)seeke;
	byte* buffer= new byte[length];
	bytesread=read(treeFile,buffer, length); 
	readTree(buffer);
}

void RBTree::insert(RBNode * & T, RBNode * pRBNode)
{
	if(pRBNode == NULL)
		return;   //nothing to insert
	if( T == NULL )  //the tree is empty
	{
		T = pRBNode;
		T->left = T->right = T->parent = NULL;
		return;
	}
	RBNode * p = T;
	RBNode * q = NULL;
	while(p != NULL)
	{
		if(pRBNode->getKey() < p->getKey())
		{
			q = p;
			p = p->left;
		}
		else
		{
			q = p;
			p = p->right;
		}
	}
	//q now points to the node after which we have to insert
	if(pRBNode->getKey() < q->getKey())
	{
		q->left = pRBNode;
		pRBNode->parent = q;
	}
	else
	{
		q->right = pRBNode;
		pRBNode->parent = q;
	}
}

void RBTree::printTree(RBNode * root, int counter)
{
	if(root == NULL)
		return;
	counter += 5;
	printTree(root->right, counter);
	for(int i = 0; i<=counter; i++)
		cout <<" ";
	char color = root->getColor() == RED ? 'r':'b';
	cout <<"("<< root->getKey() <<"|"<< color <<")\n";
	printTree(root->left, counter);
}

void RBTree::leftRot(RBNode * & root, RBNode * x)
{
	if(x->right == NULL)
		return;				//we can do no left rotation
	RBNode * y = x->right;
	x->right = y->left;
	if(y->left != NULL)
		y->left->parent = x;
	y->parent = x->parent;
	if(x->parent == NULL)
		root = y;
	else
	{
		if(x == x->parent->left)  //x is the left child of its parent
			x->parent->left = y;
		else					  //y is the right child of its parent	
			x->parent->right = y;
	}
	y->left = x;
	x->parent = y;
}

void RBTree::rightRot(RBNode * & root, RBNode * x)
{
	if(x->left == NULL)
		return;			//we cannot do a right rotation
	RBNode * y = x->left;
	x->left = y->right;
	if(y->right != NULL)
		y->right->parent = x;
	y->parent = x->parent;
	if(x->parent == NULL)
		root = y;
	else
	{
		if(x == x->parent->left)
			x->parent->left = y;
		else
			x->parent->right = y;
	}
	y->right = x;
	x->parent = y;
}

void RBTree::rbInsert(RBNode * & root, RBNode * x)
{	
	if(x == NULL)
		return;   //nothing to insert
	insert(root, x);
	x->setColor(RED);
	if(x->parent == NULL || x->parent->parent == NULL)  //we have inserted the root its child
	{
		root->setColor(BLACK);		//by convention the root is always BLACK
		return;
	}
	//the inserted node x is neither the root, nor any of its children
	while( x != root && x->parent->getColor() == RED )
	{
		if(x->parent == x->parent->parent->left) //x->parent is a left child
		{
			RBNode * y = x->parent->parent->right;  //y = the uncle of x
			COLOR colY;
			if(y == NULL)
				colY = BLACK;
			else
				colY = y->getColor();
			if(colY == RED)    //Case 1
			{
				x->parent->setColor(BLACK);
				y->setColor(BLACK);   //y cannot be NULL since it is RED
				y->parent->setColor(RED);
				x = x->parent->parent;
			}
			else  //color of y is BLACK (it could be NULL)
			{
				if(x == x->parent->right)  //Case 2
				{
					x = x->parent;
					leftRot(root, x);
				}
			//Case 3				
				x->parent->setColor(BLACK);
				x->parent->parent->setColor(RED);
				rightRot(root, x->parent->parent);				
			}
		}
		else  //symmetric case when x->parent is a right child
		{
			RBNode * y = x->parent->parent->left;			//y = the uncle of x
			COLOR colY;
			if(y == NULL)
				colY = BLACK;
			else
				colY = y->getColor();
			if(colY == RED)
			{
				y->parent->setColor(RED);  //y cannot be null since it is RED
				y->setColor(BLACK);
				x->parent->setColor(BLACK);
				x = x->parent->parent;
			}
			else   //color of Y is BLACK;
			{
				if(x == x->parent->left)  //case 2
				{
					x = x->parent;
					rightRot(root, x);
				}
				//case 3						
				x->parent->parent->setColor(RED);			
				x->parent->setColor(BLACK);				
				leftRot(root,x->parent->parent);
			}
		}
	}
	root->setColor(BLACK);
}

RBNode* RBTree::successor(RBNode * root)
{
	if(root == NULL)
		return NULL;
	if(root->right == NULL)
		return NULL;
	RBNode * p = root->right;
	while(p->left != NULL)
		p = p->left;
	return p;
}

void RBTree::rbDelete(RBNode * & root, RBNode * z)
{
	if(z == NULL)
		return;			//nothing to delete
	if(root == NULL)
		return;
	if(z == root && z->left == NULL && z->right == NULL)
	{
		z->parent = NULL;
		delete z;
		root = NULL;
		return;
	}
	//there are at least 2 nodes in the tree
	RBNode * y = NULL;
	RBNode * x = NULL;
	if(z->left == NULL || z->right == NULL)
		y = z;
	else
		y = successor(z);
	if(y->left != NULL)
		x = y->left;
	else
		x = y->right;
	if(x != NULL)
		x->parent = y->parent;
	else
	{
		x = nil;  //this is kind of a santinel, a global variable		                       
		x->parent = y->parent;
	}
	if(y->parent == NULL)
		root = x;     //this x cannot be a santinel because the root must have at least one child
	                  //the opposite was checked at the very beginning of the function
	else
		if(y == y->parent ->left)
			y->parent->left = x;
		else
			y->parent->right = x;
	if(y != z)
		z->setKey(y->getKey());
	COLOR colY = y->getColor();
	//we delete y
	y->left = y->right = y->parent = NULL;
	delete y;
	if(colY == BLACK)	//we might have violated property 4 or 3
	{
		if(x->getColor() == RED)
			x->setColor(BLACK);
		else
			rbFix(root,x);  
	}
	else  //we don't need to fix, we delete x if it is a sentinel
		if(x == nil)
		{
			if(x == x->parent->left)
				x->parent->left = NULL;
			else
				x->parent->right = NULL;
			x->parent = x->left = x->right = NULL;
			x->setColor(BLACK);   //not needed, but better be safe :)
		}
}


//helper function, used to avoid checking in the code
//returns the color of p, or BLACK if p is NULL
COLOR RBTree::colorOf(RBNode * p)
{
	if(p != NULL)
		return p->getColor();
	else
		return BLACK;
}

void RBTree::removeSentinel(RBNode * p)
{
	if(p == nil)
	{
		if(p == p->parent->left)
			p->parent->left = NULL;
		else
			p->parent->right = NULL;
		p->parent = p->left = p->right = NULL;
		p->setColor(BLACK);     //not needed
	}
}

void RBTree::rbFix(RBNode * & root, RBNode * x)
{
	RBNode * w = NULL;
	while( x != root && x->getColor() == BLACK )
	{
		if(x == x->parent->left)   //x is a left child
		{
			w = x->parent->right;
			if(w != NULL)
			{	
				if(w->getColor() == RED)  //Case1  (brother of x is RED)
				{
					w->parent->setColor(RED);  //w cannot be NULL, since it is RED
					w->setColor(BLACK);
					leftRot(root, x->parent);
					w = x->parent->right;				
				}
				else //w is BLACK => Cases 2, 3 or 4
				{
					if(colorOf(w->left) == BLACK && colorOf(w->right) == BLACK)  //case 2
					{						
						w->setColor(RED);
						RBNode * p = x;
						x = x->parent;
						x->setColor(BLACK);
						removeSentinel(p);
					}
					else  //not both children of w are BLACK => case3 or case4
					{
						if(colorOf(w->right) == BLACK)  //i.e. w->left == RED => Case 3
						{
							w->setColor(RED);
							w->left->setColor(BLACK); //w->left is RED hence it cannot be NULL
							rightRot(root, w);
						}
					    else //case4   w->right == RED
						{
							w->setColor(x->parent->getColor());
							x->parent->setColor(BLACK);
							w->right->setColor(BLACK);
							leftRot(root, x->parent);
							RBNode * p = x;
							x = root;  //we have finished
							removeSentinel(p);
						}						
					}
				}
			}
			else      //w is NULL
			{
				//set x->parent BLACK, pass x up to its parent and remove the sentinel if necessary
				x->parent->setColor(BLACK);
				RBNode * p = x;
				x = x->parent;
				removeSentinel(p);			
			}
		}
		else  //symmetric case if x is a right child
		{
			w = x->parent->left;
			if(w != NULL)
			{
				if(w->getColor() == RED) //Case 1
				{
					w->parent->setColor(RED);  //w cannot be NULL, since it is RED
					w->setColor(BLACK);
					rightRot(root, x->parent);
					w = x->parent->left;			
				}
				else  //color of w is BLACK => case 2, 3 or 4
				{
					if(colorOf(w->left) == BLACK && colorOf(w->right) == BLACK)  //case 2
					{						
						w->setColor(RED);
						RBNode * p = x;
						x = x->parent;
						x->setColor(BLACK);
						removeSentinel(p);
					}
					else  //not both children of w are BLACK => case3 or case4
					{
						if(colorOf(w->left) == BLACK)  //i.e. w->right == RED => Case 3
						{
							w->setColor(RED);
							w->right->setColor(BLACK); //w->left is RED hence it cannot be NULL
							leftRot(root, w);
						}
					    else //case4   w->left == RED
						{
							w->setColor(x->parent->getColor());
							x->parent->setColor(BLACK);
							w->left->setColor(BLACK);
							rightRot(root, x->parent);
							RBNode * p = x;
							x = root;  //we have finished
							removeSentinel(p);
						}						
					}
				}
			}
			else   //w is NULL
			{
				RBNode * p = x;
				x = x->parent;
				x->setColor(BLACK);
				removeSentinel(p);
			}
		}
	}
	root->setColor(BLACK);
}

RBNode * RBTree::search(RBNode * root, long keyToSearch)
{
	if(root == NULL)
		return NULL;
	RBNode * p = root;
	while( p != NULL)  //infinite loop
	{
		long key = p->getKey();
		if(key == keyToSearch)
			return p;
		if(keyToSearch < key)
			p = p->left;
		else
			p = p->right;
	}
	return NULL;
}

void RBTree::buildHashes(RBNode* node, byte** hash)
{
	if(node == NULL)
		return;
	
	if (node->left==NULL)
		node->lDigest=NULL;
	else
	{
		node->m_leftID=node->left->m_id;
		node->m_child++;
		buildHashes(node->left, &(node->lDigest));		
	}
	
	if (node->right==NULL)
		node->rDigest=NULL;
	else
	{
		node->m_rightID=node->right->m_id;
		node->m_child++;
		buildHashes(node->right, &(node->rDigest));		
	}

	int count=0;
	if (node->lDigest!=NULL)
		count++;
	if (node->cDigest!=NULL)
		count++;
	if (node->rDigest!=NULL)
		count++;		
	if (count>0)
	{
		byte* input = (byte*)malloc(count*DIGEST_LENGTH);
		byte* tmp=input;
		if (node->lDigest!=NULL)
		{
			memcpy(tmp, node->lDigest, DIGEST_LENGTH);
			tmp+=DIGEST_LENGTH;
		}
		if (node->cDigest!=NULL)
		{
			memcpy(tmp, node->cDigest, DIGEST_LENGTH);
			tmp+=DIGEST_LENGTH;
		}
		if (node->rDigest!=NULL)
		{
			memcpy(tmp, node->rDigest, DIGEST_LENGTH);
			tmp+=DIGEST_LENGTH;
		}
		*hash=(byte*)malloc(DIGEST_LENGTH);
		if (count>1)
		{
			node->m_tree->m_c0=clock();
			sha1(input, *hash, count*DIGEST_LENGTH);	
			node->m_tree->m_c1=clock();
			node->m_tree->m_cputime+=(double)(node->m_tree->m_c1-node->m_tree->m_c0)/(double)(CLOCKS_PER_SEC);
			node->m_tree->m_hashComputed++;

		}
		else
			memcpy(*hash, input, DIGEST_LENGTH);
		delete input;
	}
}

void RBTree::showTree()
{
	std::vector<RBNode*> nodes;
	RBNode* tmp;
	if (m_root==NULL)
		return;
	nodes.push_back(m_root);
	while(!nodes.empty())
	{
		tmp=nodes[0];
		nodes.erase(nodes.begin());
		tmp->showNode();
		if (tmp->m_leftID!=-1)
			nodes.push_back(tmp->left);
		if (tmp->m_rightID!=-1)
			nodes.push_back(tmp->right);
	}
	printf("\n root ID: %d tree digest: ", m_root->m_id);
	if (m_treeDigest!=NULL)
	{	
		unsigned long j=0;
		for (j=0; j<DIGEST_LENGTH; j++)
			printf("%02x",m_treeDigest[j]); 
		printf("\n");
	}
}

void RBTree::bulkloadTree(long* keys, long* pid, int* nodeid, byte** indigests, long size)
{
	long i;

	m_id=0;

	for (i=0;i<size;i++)
	{	
		RBNode* tmp=new RBNode(this, keys[i], pid[i], nodeid[i], indigests[i]);
		rbInsert(m_root, tmp);
	}
	m_cputime=0.0;
	m_hashComputed=0;
	buildHashes(m_root,&(m_treeDigest));
	m_rootID=m_root->m_id;
}

