// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double m_xmin, m_xmax;

	Point(double x1, double x2)
	{
		m_xmin = (x1 < x2) ? x1 : x2;
		m_xmax = (x1 > x2) ? x1 : x2;
	}
};

int main(int argc, char** argv)
{
	if (argc != 3)
	{
		cerr << "Usage: " << argv[0] << " database update_file." << endl;
		return -1;
	}

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open database file " << argv[1] << "." << endl;
		return -1;
	}

	ifstream fin1(argv[2]);
	if (! fin1)
	{
		cerr << "Cannot open update file " << argv[2] << "." << endl;
		return -1;
	}


	map<unsigned long, int> data;
	unsigned long x, y;
	int op;
	unsigned long id;
	unsigned long totalInsert=0;

	while (fin)
	{
		fin >> op >> id >> x >> y;
		if (! fin.good()) continue; // skip newlines, etc.
		data.insert(pair<unsigned long, int>(x, 0));	
	}
	
	while(fin1)
	{
		fin1 >> op >> id >> x >> y;
		if (! fin1.good()) continue; // skip newlines, etc.
		data.insert(pair<unsigned long, int>(x, 1));	
		totalInsert++;
	}
	
	map<unsigned long, int>::iterator it = data.begin();
	map<unsigned long, int>::iterator next = data.begin();
	next++;

	int status1;
	int status2;
	unsigned long pairs=0;
	for (it;it!=data.end();it++,next++)
	{
		if (next==data.end())
			break;
		status1=(*it).second;
		status2=(*next).second;
		if (status1==1 && status2==1)
			pairs++;
		//cout<<(*it).first<<endl;
	}
	unsigned long compute=2*totalInsert-pairs;
	cout<<"total insert: "<<totalInsert<<endl;
	cout<<"pairs: "<<pairs<<endl;
	cout<<"need to recompute signatures: "<<compute<<endl;
	return 0;
}
