// Spatial Index Library
//
// Copyright (C) 2002 Navel Ltd.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    lifeifei@cs.bu.edu, mhadji@gmail.com

#include <tools/Tools.h>

#include <set>

using namespace std;

#define INSERT 1
#define DELETE 0
#define QUERY 2
#define RANGE 1000000

class Point 
{
public:
	double m_xmin, m_xmax;

	Point(double x1, double x2)
	{
		m_xmin = (x1 < x2) ? x1 : x2;
		m_xmax = (x1 > x2) ? x1 : x2;
	}
};

int main(int argc, char** argv)
{
	if (argc != 4)
	{
		cerr << "Usage: " << argv[0] << " data_file number_of_query selectivity." << endl;
		return -1;
	}

	ifstream fin(argv[1]);
	if (! fin)
	{
		cerr << "Cannot open data file " << argv[1] << "." << endl;
		return -1;
	}

	unsigned long numberOfQuery = atol(argv[2]);
	double selectivity= atof(argv[3]);
	map<unsigned long, unsigned long> data;
	vector<unsigned long> keys;
	Tools::Random rnd;
	unsigned long x, y;
	int op;
	unsigned long id;

	while (fin)
	{
		fin >> op >> id >> x >> y;
		if (! fin.good()) continue; // skip newlines, etc.
		data.insert(pair<unsigned long, unsigned long>(x, id));	
		keys.push_back(x);
	}
	
	std::sort(keys.begin(),keys.end());

	unsigned long numberOfObjects=keys.size();
	unsigned long range=(unsigned long)ceil(selectivity*numberOfObjects);
	for (unsigned long i = 0; i < numberOfQuery; i++)
	{
		unsigned long start = static_cast<unsigned long>(rnd.nextUniformLong(0, numberOfObjects-range-1));
		unsigned long x1 = keys[start];
		unsigned long x2 = keys[start+range-2];
		map<unsigned long, unsigned long>::iterator it1 = data.find(x1);
		assert(it1 != data.end());
		map<unsigned long, unsigned long>::iterator it2 = data.find(x2);
		assert(it2 != data.end());
		cout << QUERY << " " << (*it1).second << " "<< x1 << " " << x2 << endl;
	}

	return 0;
}
